import 'dart:async';
import 'dart:math';

import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:path_stack/src/path_stack_path_provider.dart';
import 'package:path_to_regexp/path_to_regexp.dart';
import 'path_stack_notifications.dart';
import 'route_builder.dart';
export 'route_builder.dart';
export 'path_stack_path_provider.dart';

class PathStack extends StatefulWidget {
  static Duration defaultDuration = Duration(milliseconds: 250);
  static String defaultPath = "/";

  /// Determines which route gets shown. This is analogous to `IndexedStack.index`
  /// It is nullable because PathStack will inherit .path from any parents, and also any [PathStackPathProvider] widgets above it.
  /// If no path can be located, and no path is assigned, it will default to `PathStack.defaultPath`
  final String? path;

  /// Sets the base path for all child routes, similar to `base href` in html.toStackRoute
  /// By default, this inherits from parent Stacks, in a recursive manner so you normally don't have to set it.
  final String? basePath;

  /// A map of all routes that this path stack can handle
  final Map<List<String>, PathStackRouteBuilder> routes;

  /// Whether this stack is caseSensitive when looking for path matches
  final bool caseSensitive;

  /// Whether to output more verbose debug logs
  final bool verboseMode;

  /// Wraps all child routes in a shared widget
  final Widget Function(BuildContext context, Widget child)? scaffoldBuilder;

  /// Returns a widget to show when no path match can be found
  final PathStackRouteBuilder? unknownRouteBuilder;

  /// A callback whenever route changes. A [PathStackRouteChangingNotification] is also dispatched.
  final void Function(String path, Map<String, String> args, {String matchedRoute})? onRouteChanging;

  /// Provides a custom animation when a new route is shown. Supports only `transitionIn` right now.
  final Widget Function(BuildContext context, Widget child, AnimationController animation)? transitionBuilder;

  /// The duration of the transitionIn animation
  final Duration? transitionDuration;

  PathStack(
      {Key? key,
      this.path,
      required this.routes,
      this.basePath,
      this.caseSensitive = false,
      this.scaffoldBuilder,
      this.transitionBuilder,
      this.transitionDuration,
      this.unknownRouteBuilder,
      this.onRouteChanging,
      this.verboseMode = false})
      : super(key: key);

  static PathStackState? of(BuildContext context) {
    _InheritedPathStack? stack = context.dependOnInheritedWidgetOfExactType<_InheritedPathStack>();
    return stack?.state;
  }

  @override
  PathStackState createState() => PathStackState();
}

class PathStackState extends State<PathStack> with SingleTickerProviderStateMixin {
  // A cache of widgets, by path. Each widget is rebuilt each time it's shown, but the state is retained.
  Map<String, Widget> _knownRoutes = {};
  String? _matchedRoutePaths;

  /// Controls the transitionIn animation if there is one (there is no animation by default)
  late AnimationController transitionInAnim =
      AnimationController(vsync: this, duration: widget.transitionDuration ?? PathStack.defaultDuration)..forward();

  /// Contains the last known args so any child can look them up with `PathStack.of(context).args`
  String _oldPath = "";
  Map<String, String> _args = {};
  Map<String, String> get args => _args;

  /// Uses widget.basePath -> currentRoutingPath
  String get basePath => widget.basePath ?? PathStack.of(context)?.currentRoutingPath ?? "";

  /// Uses widget.path ->  inherited PathStack -> inherited PathStackPathProvider -> [PathStack.defaultPath]
  String get path =>
      (widget.path ?? PathStack.of(context)?.path ?? PathStackPathProvider.of(context)?.path ?? PathStack.defaultPath);

  // Gets the route of this stacks, and all parent stacks.
  // This is used as the `basePath` when none is specified so all child routes will be relative to the parents location.
  String get currentRoutingPath =>
      (PathStack.of(context)?.currentRoutingPath ?? "") + (_matchedRoutePaths?.split(",")[0] ?? "");

  @override
  Widget build(BuildContext context) {
    // Try and find a known route for the current path
    _RoutePathWithBuilder routeToShow = getRouteBuilderForPath(path);

    // Extract queryString and path args
    _args = getArgsFromPath(route: basePath + routeToShow.path, path: path);

    // Create widget, so we can access any config it might have
    Widget routeWidget = routeToShow.content.call(this);

    /// Wrap widget in a [RouteConfig] if it's not already using one.
    if (widget.verboseMode) {
      print("Build path = $path");
    }
    RouteConfig configuredRoute = (routeWidget is RouteConfig)
        ? routeWidget
        : RouteConfig(
            child: routeWidget,
          );
    // Check onBeforeChange, if  this returns false we still need to finish this method, but we should make no changes.
    bool allowChange = configuredRoute.onBeforeEnter?.call(routeToShow.path) ?? true;
    if (allowChange && _oldPath != path) {
      // Change the path, and trigger events/animations associated with that
      _changeRoute(routeToShow.path, configuredRoute);
      _oldPath = path;
    }

    /// Create IndexedStack
    Widget content = _buildIndexedStackFromCache();

    /// Wrap builders
    content = widget.transitionBuilder?.call(context, content, transitionInAnim) ?? content;
    content = widget.scaffoldBuilder?.call(context, content) ?? content;

    /// Wrap content with inherited widget so we can look it up
    return _InheritedPathStack(child: content, state: this);
  }

  void _changeRoute(String newRoutePaths, RouteConfig child, {bool force = false}) {
    // Inform listeners we're about to change
    widget.onRouteChanging?.call(newRoutePaths, args, matchedRoute: newRoutePaths);
    // Use notification to allow ancestors to listen
    PathStackRouteChangingNotification(newRoutePaths, args, matchedRoute: newRoutePaths).dispatch(context);
    // Trigger transition
    transitionInAnim.forward(from: 0);
    _matchedRoutePaths = newRoutePaths;
    // Add or update the item in the cache
    print("Changing route, _matchedRoutePaths = $_matchedRoutePaths");
    _knownRoutes[newRoutePaths] = child;
    // Remove persistent routes next frame, this way we can keep things simple during this build method.
    if (child.maintainState == false) _removeRouteNextFrame(newRoutePaths);
  }

  Widget _buildIndexedStackFromCache() {
    // Get a list of the children from our known cache
    List<Widget> children = _knownRoutes.values.toList(); // Make a copy of the current knownChildren
    // Get the index for the current child, based on string path
    int index = _knownRoutes.keys.toList().indexWhere((r) => r == _matchedRoutePaths);
    return IndexedStack(index: index, children: children);
  }

  /// Clears all known routes from memory
  void clearState({bool includeCurrent = true}) {
    Widget? currentRoute = _knownRoutes[_matchedRoutePaths];
    _knownRoutes.clear();
    if (currentRoute != null) {
      _knownRoutes[_matchedRoutePaths!] = currentRoute;
    }
  }

  @override
  void dispose() {
    clearState();
    super.dispose();
  }

  /// Return true if the path matches any of the routes
  bool _checkPathAgainstRoutes(String path, List<String> routes, {bool? exactMatch}) {
    bool allowSuffix;
    // Use exact match if it's been specified
    if (exactMatch != null) {
      allowSuffix = !exactMatch;
      // Otherwise fallback to auto-behavior
    } else {
      // The first entry in the list of routes will determine whether it allowsSuffix or not, make a special exception for "/" which is always exact match.
      allowSuffix = routes[0].endsWith("/") && routes[0] != "/";
    }
    // Use pathToRegExp to check the paths for matches
    for (var i = 0; i < routes.length; i++) {
      final regExp = pathToRegExp('$basePath${routes[i]}', prefix: allowSuffix, caseSensitive: widget.caseSensitive);
      if (regExp.hasMatch("$path")) return true;
    }
    return false;
  }

  /// Given a path, this finds a matching route builder, or it shows an "Unknown Page" Widget
  _RoutePathWithBuilder getRouteBuilderForPath(String path) {
    _RoutePathWithBuilder? matchingRoute;
    // For each route, check all paths it has defined and see if we have a match
    widget.routes.forEach((key, value) {
      if (_checkPathAgainstRoutes(path, key) && matchingRoute == null) {
        matchingRoute = _RoutePathWithBuilder(key.join(","), value);
      }
    });
    // If unable to find route, fallback to unknownPathBuilder or a default 404 page.
    if (matchingRoute == null) {
      print("WARNING: Unable to find a route for $path");
      matchingRoute = _buildUnknownPageWidget();
    } else {
      //print("RouteMatched, ${matchingRoute?.path}");
    }
    return matchingRoute!;
  }

  /// TODO: This could be extracted
  /// Extracts a Map<String,String> from a path
  Map<String, String> getArgsFromPath({required String route, required String path}) {
    // Get path params
    List<String> paramList = [];
    final regExp = pathToRegExp(route, parameters: paramList);
    Map<String, String> pathMap = {};
    if (regExp.hasMatch(path)) {
      final match = regExp.matchAsPrefix(path)!;
      pathMap = extract(paramList, match);
    }
    // Get queryString params
    Map<String, String> queryMap = Uri.parse(path).queryParameters;
    // Combine path and query args into one, the query args will override the path items with the same id
    final o = {...pathMap, ...queryMap};
    return o;
  }

  /// Removes a route from the cache. Used with non-persistent routes.
  /// We do it this way primarily to simplify the code within build()
  void _removeRouteNextFrame(String routePaths) => scheduleMicrotask(() => _knownRoutes.remove(routePaths));

  /// This could be extracted to a dedicated widget
  _RoutePathWithBuilder? _buildUnknownPageWidget() {
    return _RoutePathWithBuilder("404", (info) {
      return widget.unknownRouteBuilder?.call(info) ??
          Scaffold(
            body: Center(
              child: Column(
                mainAxisSize: MainAxisSize.min,
                children: [
                  Text("Page Not Found :(", style: TextStyle(fontSize: 26)),
                  Text("Unable to match a route to: '$path'"),
                ],
              ),
            ),
          );
    });
  }
}

// Holds a typed version of the PathStack.routes map
class _RoutePathWithBuilder {
  _RoutePathWithBuilder(this.path, this.content);
  final String path; // concatenated list of paths for their route
  final PathStackRouteBuilder content;
}

class _InheritedPathStack extends InheritedWidget {
  _InheritedPathStack({Key? key, required Widget child, required this.state}) : super(key: key, child: child);
  final PathStackState state;
  @override
  bool updateShouldNotify(covariant InheritedWidget oldWidget) => true;
}
